%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                    The CIFF Proof Procedure                     %%%
%%%        for Abductive Logic Programming with Constraints         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Proof procedure (file ciff-proc.pl)                             %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v2.0: 03/09/2004 by Ulle Endriss (ue@doc.ic.ac.uk)              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This file provides the implementation of the actual proof 
%%% procedure. The core predicate is sat/7. The second most important
%%% predicate is simplify_equalities/4, which implements the rewrite
%%% rules for equalities. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% sat( +Node, +EV, +CL, +LM, +Defs, +FreeVars, -Answer).
%%% Apply rules of the CIFF procedure to the given Node as long
%%% as possible (i.e. saturate), using the definitions in Defs, and
%%% return an answer. The set of all answers can be collected via 
%%% forced backtracking.
%%%
%%% Node is a list of goals, representing a conjunction. From a 
%%% syntactic point of view, we can distinguish three kinds of goals:
%%%  - atoms
%%%  - Cs implies Ds, where Cs is a list of literals, representing a
%%%    conjunction, and Ds is lists of atoms, representing a
%%%    disjunction (this includes the case of single negative literals)
%%%  - lists of lists of either of the above, representing disjunctions 
%%%    of conjunctions of formulas
%%% Node should initially be set to the list of goals in the query 
%%% together with the list of integrity constraints.
%%%
%%% FreeVars is the list of free variables, i.e. the list of variables
%%% appearing in the original query. They are only needed during answer 
%%% extraction at the very end.
%%%
%%% EV (for existentially quantified variables) is the set of 
%%% existentially quantified variables (including any free variables) 
%%% for the respective node. Strictly speaking, this list will not 
%%% always include all existentially quantified variables, but only 
%%% those for which we may need to check the type of quantification in 
%%% order to determine whether or not a particular rule is applicable.
%%% In particular, variables that only occur inside a disjunction
%%% (within a node) to which the splitting rule is yet to be applied,
%%% need not be classified as either existentially or universally 
%%% quantified. EV is updated whenever a branch is split. 
%%% It should initially be set to the list of free variables.
%%%
%%% CL (for constraint list) is the list of CLP constraints that have 
%%% so far been encountered within the respective node. In fact, CL is
%%% a pair of lists CL1:CL2. CL1 is the list of constraints and CL2 is
%%% a copy of that list. The constraints actually posted to the 
%%% constraint solver are taken from CL2, while CL1 is used to report
%%% answers. This avoids variables in the answer being instantiated,
%%% i.e. we can report substitutions and constraints on the meta-level.
%%% Initially, CL should be set to []:[].
%%%
%%% LM (for loop management) is a list of expressions of the form A:B 
%%% recording pairs of formulas that have already been used for 
%%% propagation or factoring for that node (to be able to avoid loops).
%%% LM should be initialised with the empty list.
%%%
%%% Note that the procedure may loop in case there are cyclic 
%%% definitions in Defs.
%%%
%%% This implementation does not cover the two additional splitting
%%% rules (for implications) suggested in the original paper by Fung 
%%% and Kowalski on the IFF procedure (Journal of Logic Programming 
%%% 33(2):151-165, 1997). The intention behind this change is to avoid 
%%% generating answers that are not minimal. To include the original 
%%% splitting rules, you could add clauses that rewrite implications as 
%%% follows (the rest will be done by the basic splitting rule):
%%%   (1) A -> B v Bs = (A -> B) v Bs  unless Bs = []
%%%   (2) A -> B = (A -> false) v B    unless B = false
%%%
%%% The current implementation does not attempt to remove duplicate
%%% (or otherwise subsumed) subgoals.
%%%
%%% In principle, the order of clauses in the definition of sat/7 may 
%%% be changed freely (apart from the last two). This may help to find 
%%% out (by experiment) what rule application strategy yields the 
%%% shortest derivations. However, note that some orders may be more 
%%% likely to help avoid looping than others.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% removing failure nodes [false,...] (including empty disjunction []):
sat( Node, _, _, _, _, _, _) :-
  (member( false, Node) ; member( [], Node)), 
  inform( 'removed failure node', false, [false]), !,
  fail.

% logical simplification (redundant formulas): 
% remove [false,...] implies A
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies B, Node),
  member( false, As), 
  delete( Node, As implies B, NewNode),
  inform( 'simplification (remove [false,...] -> A)', As implies B, NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% logical simplification (trivial condition): 
% rewrite [] -> A (representing true -> A) as A
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( [] implies As, Node), 
  delete( Node, [] implies As, Node1),
  listify( As, As1),
  add_disjunction( Node1, EV, As1, NewNode, NewEV),
  inform( 'simplification ([] -> A)', [] implies As, NewNode), !,
  sat( NewNode, NewEV, CL, LM, Defs, FreeVars, Answer).

% negation rewriting: 
% rewrite (A -> false) ^ B -> C as B -> A v C
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies Bs, Node),
  \+ (member( Imp:_, LM), Imp == As implies Bs),
  member( [A] implies [false], As), 
  delete( Node, As implies Bs, Node1),
  delete( As, [A] implies [false], As1),
  NewNode = [As1 implies [A|Bs]|Node1],
  inform( 'negation rewriting', As implies Bs, NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% equality rewriting for atoms:
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( T1=T2, Node),
  simplify_equalities( [T1=T2], EV, Equalities),
  \+ ([T1=T2] == Equalities), %  to avoid loops
  delete( Node, T1=T2, Node1),
  append( Node1, Equalities, NewNode),
  inform( 'equality rewriting for atom', T1=T2, NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% substitution rule for atoms: 
% retain X/T instead of X=T (to avoid loops) for answer substitution
% replace by the corresponding constraint in case this is admissible
sat( Node, EV, CL:CCL, LM, Defs, FreeVars, Answer) :-
  member( X=T, Node),
  var( X),
  \+ (subterm( X, T) ; X == T), 
  delete( Node, X=T, Node1),
  subst( X/T, Node1*LM, Node2*NewLM),
  term_variables( CL, CVars),
  ((exact_member( X, CVars) ; exact_member( T, CVars)) -> 
    equality_constraint( X, T, C),
    NewNode = [C|Node2] 
  ; 
    NewNode = [X/T|Node2]
  ),
  inform( 'substitution rule for atom', X=T, NewNode), !,
  sat( NewNode, EV, CL:CCL, NewLM, Defs, FreeVars, Answer).

% equality rewriting for implications:
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies B, Node),
  \+ (member( Imp:_, LM), Imp == As implies B),
  member( T1=T2, As),
  simplify_equalities( [T1=T2], EV, Equalities),
  \+ ([T1=T2] == Equalities), %  to avoid loops
  delete( Node, As implies B, Node1),
  delete( As, T1=T2, As1),
  append( As1, Equalities, As2),
  NewNode = [As2 implies B|Node1],
  inform( 'equality rewriting for implication', T1=T2, NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% substitution rule for implications:
% Given A -> B with X=T within A  
% apply provided X is universally quantified, X does not occur within T,
% and either T contains no universally quantified variables or 
% X does not occur within B
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies B, Node),
  \+ (member( Imp:_, LM), Imp == As implies B),
  member( X=T, As),
  var( X),
  \+ exact_member(X, EV), 
  \+ (subterm( X, T) ; X == T),
  term_variables( T, Vars),
  (exact_subset( Vars, EV) ; \+ subterm( X, B)),
  delete( Node, As implies B, Node1),
  delete( As, X=T, As1),
  subst( X/T, As1 implies B, Implication), 
  NewNode = [Implication|Node1],
  inform( 'substitution rule for implication', As implies B, NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% constraint solving:
% move constraints into CL and perform a (simple) consistency check
sat( Node, EV, CL:CCL, LM, Defs, FreeVars, Answer) :-
  member( C, Node), 
  is_constraint( C), 
  copy_term( C:CL, CC:CCL), !,
  (post_constraint( CC) ->
    delete( Node, C, NewNode),
    inform( 'constraint checking (success)', [C]:CL, NewNode),
    ((ground( C) ; exact_member( C, CL)) -> NewCL = CL:CCL ; NewCL = [C|CL]:[CC|CCL]), !,
    sat( NewNode, EV, NewCL, LM, Defs, FreeVars, Answer)
  ;
    inform( 'constraint checking (failure)', [C]:CL, [false]),
    fail
  ).

% unfolding atoms:
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( A, Node),
  is_atom( A),
  get_def( A, Defs, Ds), % fails if A is abducible
  delete( Node, A, Node1),
  add_disjunction( Node1, EV, Ds, NewNode, NewEV),
  inform( 'unfolding atom', A iff Ds, NewNode), !,
  sat( NewNode, NewEV, CL, LM, Defs, FreeVars, Answer).

% unfolding within implications:
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies B, Node),
  \+ (member( Imp:_, LM), Imp == As implies B),
  unfoldable_implication( As implies B), 
  member( A, As),
  get_def( A, Defs, Ds), % fails if A is abducible
  delete( Node, As implies B, Node1),
  delete( As, A, As1),
  unfold_implication( As1 implies B, Ds, Imps),
  append( Imps, Node1, NewNode),
  inform( 'unfold within implication', (As implies B):(A iff Ds), NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% case analysis for equalities: 
% rewrite (X=T ^ A) -> B as (X=T ^ (A -> B)) v (X=T -> false)
% provided X is existentially quantified, T does not contain X, 
% and T is not a universally quantified variable
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies B, Node),
  \+ (member( Imp:_, LM), Imp == As implies B),
  member( X=T, As),
  var( X),
  exact_member( X, EV),
  \+ (subterm( X, T) ; X == T),
  (var( T) -> exact_member( T, EV) ; true),
  \+ (As implies B = [X=T] implies [false]), % to avoid loops
  delete( Node, As implies B, Node1),
  delete( As, X=T, As1),
  D = [[X=T, As1 implies B], [[X=T] implies [false]]],
  NewNode = [D|Node1],
  inform( 'case analysis', (As implies B):(X=T), NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% case analysis for constraints: 
% rewrite (C ^ A) -> B as [C ^ (A -> B)] v C' 
% where C is a constraint and C' is its complement
% provided all variables occurring in C are existentially quantified
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies B, Node),
  \+ (member( Imp:_, LM), Imp == As implies B),
  member( C, As),
  is_constraint( C),
  term_variables( C, Vars),
  exact_subset( Vars, EV), 
  delete( Node, As implies B, Node1),
  delete( As, C, As1),
  complementary_constraint( C, C1),
  D = [[C, As1 implies B], [C1]],
  NewNode = [D|Node1],
  inform( 'case analysis for constraints', (As implies B):C, NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).
  
% factoring: 
% rewrite P(t1..tn) ^ P(s1..sn) as 
% (P(t1..tn) ^ P(s1..sn) ^ (t1=s1 ^ ... ^ tn=sn) -> false) v
% (P(t_1..tn) ^ t1=s1 ^ ... ^ tn=sn), provided that P is abducible
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :- 
  factoring( on),
  member( A1, Node),
  is_atom( A1),
  A1 =.. [P|Ts1],
  member( A2, Node),
  is_atom( A2),
  A2 =.. [P|Ts2],
  same_length( Ts1, Ts2),
  \+ Ts1 == Ts2,
  is_abducible( A1, Defs),
  \+ (exact_member( A1:A2, LM) ; exact_member( A2:A1, LM)), 
  NewLM = [A1:A2|LM],
  delete( Node, A1, Node1),
  delete( Node1, A2, Node2),
  construct_equalities( Ts1, Ts2, Equalities),
  D = [[A1, A2, Equalities implies [false]], [A1|Equalities]],
  NewNode = [D|Node2],
  inform( 'factoring', A1:A2, NewNode), !,
  sat( NewNode, EV, CL, NewLM, Defs, FreeVars, Answer).

% rewriting unary disjunctions
% redundant: this rule has been integrated into the rules for 
% unfolding atoms as well as rewriting of [] -> A 
%sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
%  member( [Conjunction], Node), 
%  delete( Node, [Conjunction], Node1),
%  append( Conjunction, Node1, NewNode),
%  get_atoms_in_list( Conjunction, Atoms),
%  term_variables( EV:Atoms, NewEV),
%  inform( 'rewriting unary disjunction', [Conjunction], NewNode), !,
%  sat( NewNode, NewEV, CL, LM, Defs, FreeVars, Answer).

% splitting:
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( [D|Ds], Node), % a list inside Node represents a disjunction
  delete( Node, [D|Ds], Node1), 
  split( [D|Ds], Node1, EV, Choices), !, % backtracking over branches
  member( NewNode:NewEV, Choices), 
  inform( 'splitting', [D|Ds], NewNode), 
  sat( NewNode, NewEV, CL, LM, Defs, FreeVars, Answer).

% propagation:
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies B, Node),
  propagatable( A, As),
  \+ is_special_atom( A),
  A =.. [P|Ts],
  member( A1, Node),
  A1 =.. [P|Ts1],
  same_length( Ts, Ts1),
  \+ exact_member( (As implies B):A1, LM), 
  NewLM = [(As implies B):A1|LM],
  delete( As, A, As1),
  construct_equalities( Ts, Ts1, Equalities),
  append( As1, Equalities, As2),
  NewNode = [As2 implies B|Node],
  inform( 'propagation', (As implies B):A1, NewNode), !,
  sat( NewNode, EV, CL, NewLM, Defs, FreeVars, Answer).

% logical simplification (trivial condition):
% remove true in antecedent
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( As implies B, Node), 
  \+ (member( Imp:_, LM), Imp == As implies B),
  member( true, As), 
  delete( Node, As implies B, Node1),
  delete( As, true, As1),
  NewNode = [As1 implies B|Node1],
  inform( 'simplification (remove true from antecedent)', As implies B, NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% logical simplification (redundant formulas): 
% remove true (occuring as an atom)
sat( Node, EV, CL, LM, Defs, FreeVars, Answer) :-
  member( true, Node), 
  delete( Node, true, NewNode),
  inform( 'simplification (remove atom true)', true, NewNode), !,
  sat( NewNode, EV, CL, LM, Defs, FreeVars, Answer).

% dynamic allowedness rule:
% label any node as undefined that contains A -> B where A is a 
% conjunction of equalities and constraints (exception: A -> B is 
% a disequality) and no other rule applies to A -> B
sat( Node, EV, _, _, _, _, undefined) :-
  member( As implies Bs, Node),
  all_constraint_equality( As),
  \+ represents_disequality( As implies Bs, EV), 
  inform( 'dynamic allowedness rule', As implies Bs, undefined), !.

% no more rules apply:
% includes application of dis/equality-constraint rewriting
sat( Node, _, CL:CCL, _, Defs, FreeVars, Atoms:Subst:Constraints) :- 
  extract_abducible_atoms( Node, Defs, Atoms), 
  extract_substitution( Node, FullSubst),
  term_variables( CL, CLVars),
  get_subst_constraints( FullSubst, CLVars, ExtraCL),
  term_variables( FreeVars:Atoms, Vars),
  term_variables( CLVars:ExtraCL, CLVars1),
  delete_list( Vars, CLVars1, RelevantVars),
  restrict_substitution( FullSubst, RelevantVars, Subst),
  copy_term( ExtraCL:CL, ExtraCCL:CCL),
  append( ExtraCL, CL, FullCL),
  simplify_constraints( FullCL, Constraints),
  term_variables( ExtraCCL:CCL, FullCCLVars), 
  ((post_constraints( ExtraCCL), exhaustive_check( FullCCLVars)) ->
    inform( 'final constraint checking (success)', Constraints, Node)
  ;
    inform( 'final constraint checking (failure)', Constraints, [false]),
    fail
  ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% simplify_equalities( +Eqs, +EV, -NewEqs).
%%% Simplify the equalities in the list Eqs using rewrite rules 
%%% familiar from unification algorithms and transform them into a 
%%% normal form (with universally quantified variables on the lefthand 
%%% side). Reference to the list EV of existentially quantified 
%%% variables is required to be able to determine whether a variable 
%%% is universally quantified or not. 
%%%
%%% Initially, Eqs will be a list with just a single equality, but 
%%% during the rewrite process a proper list of equalities may be 
%%% generated.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% f(t1,...,tn) = g(s1,...,sm) is never true (case 2a)
simplify_equalities( [T1=T2|_], _, [false]) :-
  T1 =** [F1|_],
  T2 =** [F2|_],
  \+ (F1 = F2), !.

% f(t1,...tn) = g(s_1,...,sm) is never true for n \= m (case 2b)
simplify_equalities( [T1=T2|_], _, [false]) :-
  T1 =** [_|Ts1],
  T2 =** [_|Ts2],
  \+ same_length( Ts1, Ts2), !.

% occurs check (case 4)
simplify_equalities( [X=T|_], _, [false]) :-
  subterm( X, T), !.

% t = t is true (case 3)
simplify_equalities( [T1=T2|Rest], EV, Rest1) :-
  T1 == T2, !,
  simplify_equalities( Rest, EV, Rest1).

% rewrite f(t1,...,tn) = f(s1,...,sn) as t1=s1 ^ ... ^ tn=sn (case 1)
simplify_equalities( [T1=T2|Rest], EV, List1) :-
  T1 =** [F|Ts1],
  T2 =** [F|Ts2],
  same_length( Ts1, Ts2), 
  construct_equalities( Ts1, Ts2, Equalities),
  append( Equalities, Rest, List), !,
  simplify_equalities( List, EV, List1).

% rewrite t = X as X = t (case 5)
simplify_equalities( [T=X|Rest], EV, List) :-
  var( X),
  \+ var( T), !,
  simplify_equalities( [X=T|Rest], EV, List).

% rewrite Y = X as X = Y 
% if X is universally quantified and Y is not (case 6)
simplify_equalities( [Y=X|Rest], EV, [X=Y|List]) :-
  var( X),
  var( Y),
  \+ exact_member( X, EV),
  exact_member( Y, EV), !,
  simplify_equalities( Rest, EV, List).

% no more simplifications possible
simplify_equalities( [Head|Tail], EV, [Head|Tail1]) :-
  simplify_equalities( Tail, EV, Tail1).

% base case
simplify_equalities( [], _, []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% is_atom( +Formula).
%%% Succeed if Formula represents an atomic formula. Note that we
%%% have three kinds of basic formulas to consider: atoms, 
%%% implications, and disjunctions (represented as lists of lists).
%%% Hence, any term that is neither a list nor has "implies" as its
%%% principal functor will represent an atomic formula.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_atom( A) :-
  \+ (A = [_|_]),
  \+ (A = _ implies _).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% is_special_atom( +Atom).
%%% Succeed if Atom is a special atom such as X = Y.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_special_atom( A) :-
  functor( A, =, 2).

is_special_atom( A) :-
  functor( A, /, 2).

is_special_atom( A) :-
  is_constraint( A).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% is_abducible( +Atom, +Defs).
%%% Succeed if Atom is an instance of an abducible predicate, i.e. if
%%% it is not defined in Defs. The predicate will also fail in case
%%% Atom is a special atom (e.g. an equality).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_abducible( A, Defs) :-
  \+ member( A iff _, Defs),
  \+ is_special_atom( A).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% get_def( +Atom, +Defs, -Disjunction).
%%% Given a list of iff-definitions Defs and an Atom, return the
%%% definition for that atom (which will be a disjunction). 
%%% The variables occurring on the lefthand side of the respective
%%% iff-definition need to be substituted with the appropriate terms
%%% from Atom. Finally, any variables only occurring on the righthand
%%% side of the definition need to be renamed ("standardising apart").
%%% This predicate will fail if there is no definition for Atom in
%%% Defs, in particular in case Atom is an abducible predicate.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_def( A, Defs, Ds) :-
  A =.. [P|Ts],
  member( A1 iff Ds1, Defs),
  A1 =.. [P|Xs],
  same_length( Xs, Ts), !,
  subst( Xs, Ts, Ds1, Ds2),
  term_variables( A, Vars),
  copy_term( Ds2:Vars, Ds:Vars).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% unfold_implication( +Implication, +Disjunction, -Implications).
%%% Given an Implication and a Disjunction, construct the list of
%%% Implications we get by adding D to the list of conjuncts in the
%%% antcedent of the Implication for every D in the Disjunction.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unfold_implication( As implies B, [D|Ds], [Imp|Imps]) :-
  append( As, D, As1),
  Imp = As1 implies B,
  unfold_implication( As implies B, Ds, Imps).

unfold_implication( _, [], []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% split( +Ds, +Node, +EV, -Choices). 
%%% Distribute Node over the elements of Ds to obtain a list of new 
%%% nodes. Choices is a list of elements of the form Node1:EV1, where 
%%% Node1 is one of the new nodes and EV1 is the appropriate list of
%%% existentially quantified variables. EV1 is the list EV of 
%%% previously existentially quantified variables together with the
%%% variables occurring in atoms in the disjunct that has given rise
%%% to the new node Node1.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

split( [], _, _, []).

split( [D|Ds], Node, EV, [Node1:EV1|Choices]) :-
  append( D, Node, Node1),
  get_atoms_in_list( D, Atoms),
  term_variables( EV:Atoms, EV1),
  split( Ds, Node, EV, Choices).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% add_disjunction( +Node, +EV, +Disjunction, -NewNode, -NewEV).
%%% Add a given Disjunction to a given Node and flatten the disjunction
%%% in case it only has a single disjunct. In that case, also EV needs
%%% to be updated.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

add_disjunction( Node, EV, [Conjunction], NewNode, NewEV) :- !,
  append( Conjunction, Node, NewNode),
  get_atoms_in_list( Conjunction, Atoms),
  term_variables( EV:Atoms, NewEV).

add_disjunction( Node, EV, Disjunction, [Disjunction|Node], EV).
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% represents_disequality( +Formula, +EV).
%%% Succeed if Formulas represents a (well-defined) disequality. 
%%% This is the case if it is of the form T1=T2 -> false and any 
%%% variables in both T1 and T2 are existentially quantified.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

represents_disequality( [T1=T2] implies [false], EV) :-
  term_variables( T1:T2, Vars),
  exact_subset( Vars, EV).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% all_constraint_equality( +List).
%%% Succeed if List is a list of equalities and constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

all_constraint_equality( []).

all_constraint_equality( [_=_|Tail]) :- !,
  all_constraint_equality( Tail).

all_constraint_equality( [Head|Tail]) :-
  is_constraint( Head),
  all_constraint_equality( Tail).
  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% get_atoms_in_list( +List, -Atoms).
%%% Get the elements from list for which is_atom/1 succeeds.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_atoms_in_list( [], []).

get_atoms_in_list( [Atom|List], [Atom|Atoms]) :-
  is_atom( Atom), !,
  get_atoms_in_list( List, Atoms).

get_atoms_in_list( [_|List], Atoms) :-
  get_atoms_in_list( List, Atoms).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% extract_abducible_atoms( +List, -Atoms).
%%% Extract the abducible atoms from a given list of formulas 
%%% (a leaf node).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

extract_abducible_atoms( [], _, []).

extract_abducible_atoms( [A|Leaf], Defs, [A|As]) :-
  is_atom( A),
  is_abducible( A, Defs), !,
  extract_abducible_atoms( Leaf, Defs, As).

extract_abducible_atoms( [_|Leaf], Defs, As) :-
  extract_abducible_atoms( Leaf, Defs, As).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% extract_substitution( +List, -Substitution).
%%% Extract the variable substitutions from a given list of formulas
%%% (a leaf node). Positive substitutions are of the form X/T and 
%%% need to be collected. Negative substitution occur in the form
%%% [X=T] implies [false] and will be collected as not(X/T).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

extract_substitution( [],[]).

extract_substitution( [X/T|Leaf], [X/T|Subst]) :- !,
  extract_substitution( Leaf, Subst).

extract_substitution( [[X=T] implies [false]|Leaf], [not(X/T)|Subst]) :- !,
  extract_substitution( Leaf, Subst).

extract_substitution( [_|Leaf], Subst) :- 
  extract_substitution( Leaf, Subst).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% get_subst_constraints( +List, +ConstraintVars, -Constraints).
%%% Given a list of variable substitutions of the form X/T or not(X/T)
%%% and a list of variables that appear in constraints, construct the
%%% list of equality and disequality constraints corresponding to the
%%% substitutions.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_subst_constraints( [], _, []).

get_subst_constraints( [X/T|Subst], CVars, [C|Cs]) :-  
  (exact_member( X, CVars) ; exact_member( T, CVars)), !,
  equality_constraint( X, T, C),
  get_subst_constraints( Subst, CVars, Cs).

get_subst_constraints( [not(X/T)|Subst], CVars, [C|Cs]) :-
  (exact_member( X, CVars) ; exact_member( T, CVars)), !,
  disequality_constraint( X, T, C),
  get_subst_constraints( Subst, CVars, Cs).

get_subst_constraints( [_|Subst], CVars, Cs) :-
  get_subst_constraints( Subst, CVars, Cs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% restrict_substitution( +List, +Vars, -Substitution).
%%% Given a list of variable substitutions and a list of variables Vars,
%%% return the sublist of substitutions referring to variables in Vars.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

restrict_substitution( [], _, []).

restrict_substitution( [X/T|Subst], Vars, [X/T|ReSubst]) :-
  (exact_member( X, Vars) ; exact_member( T, Vars)), !,
  restrict_substitution( Subst, Vars, ReSubst).

restrict_substitution( [not(X/T)|Subst], Vars, [not(X/T)|ReSubst]) :-
  (exact_member( X, Vars) ; exact_member( T, Vars)), !,
  restrict_substitution( Subst, Vars, ReSubst).

restrict_substitution( [_|Subst], Vars, ReSubst) :-
  restrict_substitution( Subst, Vars, ReSubst).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% switch_factoring( +Status).
%%% Switch the factoring rule "on" or "off". The factoring rule is 
%%% not required to ensure soundness of the procedure. Using it will 
%%% result in more and more specific abductive answers being 
%%% generated (by distinguishing cases where particular abducibles 
%%% are either true or false). The default is "on".
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switch_factoring( Status) :-
  retractall( factoring(_)),
  assert( factoring(Status)).

:- switch_factoring( on). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% switch_propagation( +Type).
%%% Switch between the two variants of the propagation rule. Possible
%%% arguments are "normal" and "head". If "head" is chosen, then only
%%% propagation with the first non-constraint/non-equality atom in
%%% the antecedent of the implication is possible. Both variants are 
%%% sound and completeness is not affected either. We would expect the
%%% "head" variant to be more efficient in most cases. The default is
%%% "head".
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switch_propagation( Type) :-
  retractall( propagation(_)),
  assert( propagation(Type)).

:- switch_propagation( head). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% propagatable( +Atom, +Literalss).
%%% Check weather Atom can be propagated with an implication where the
%%% list Literals represents the conjunction of literals in the 
%%% antecedent. If the "normal" propagation rule is used, this will 
%%% succeed in case Atom matches any member in Literals; if the "head"
%%% propagation rule is used this will only succeed if Atom matches
%%% the first non-constraint/non-equality element in Literals.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

propagatable( Atom, Literals) :-
  propagation( normal),
  member( Atom, Literals).

propagatable( Atom, Literals) :-
  propagation( head),
  non_special_head( Atom, Literals).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% non_special_head( +Atom, +List).
%%% Succeed if Atom is the first element in List that is not a special 
%%% atom.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

non_special_head( Atom, [Atom|_]) :-
  \+  is_special_atom( Atom).

non_special_head( Atom, [Special|Rest]) :-
  is_special_atom( Special), 
  non_special_head( Atom, Rest).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% inform( +Rule, +Arguments, +Node). 
%%% Used for debugging purposes. The output depends on the debugging
%%% mode chosen (via switch_debugging/1). The following modes are
%%% available:
%%% "on":    full debugging information, namely print out last Rule 
%%%          applied (with Arguments) and current Node;
%%% "step":  same as "on", but Prolog will wait for the user to hit 
%%%          the return key;
%%% "short": only show the name of the Rule used;
%%% "dots":  only print a single dot for each step;
%%% "stats:  print out statistics at the end (counting takes place
%%%          within inform/3).
%%% "off":   no output [default].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

inform( _, _, _) :-
  debugging( off), !.

inform( Rule, Arguments, Node) :-
  (debugging( on) ; debugging( step)), !,
  nl, write( Rule), write( ': '), write( Arguments),
  nl, write( 'node: '), write( Node),
  (debugging( step) -> get0( _) ; nl).

inform( Rule, _, _) :-
  debugging( short), !,
  nl, write( Rule).

inform( Rule, _, _) :-
  debugging( stats), !, 
  update_stats( Rule).
  
inform( _, _, _) :-
  write( '.').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% update_stats( +Rule).
%%% Update the counters used to collect proof statistics. The action
%%% taken depends on the proof rule applied.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

update_stats( 'initialise') :- !,
  init_counter( proof_steps),
  init_counter( branches),
  init_clock.

update_stats( 'splitting') :- !,
  step_counter( proof_steps),
  step_counter( branches),
  restart_clock.

update_stats( 'final constraint checking (success)') :- !,
  step_counter( proof_steps),
  print_stats.

update_stats( _) :-
  step_counter( proof_steps).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% print_stats.
%%% Print out statistics (in case debugging mode "stats" is in use).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

print_stats :- !,
  stop_clock,
  nl,
  write( 'CIFF Statistics'), nl,
  write( '---------------'), nl,
  write( 'Steps:     '), 
  read_counter( proof_steps, ProofSteps),
  write( ProofSteps), nl,
  write( 'Branches:  '),
  read_counter( branches, BranchesCounted),
  (BranchesCounted = 0 -> Branches = 1 ; Branches = BranchesCounted),
  write( Branches), nl,
  write( 'Time [ms]: '),
  read_clock( Time),
  write( Time), nl, nl, !.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Assert that the unfolding rule may be applied to any implication.
%%% A specific application may overwrite the definition of the 
%%% predicate unfoldable_implication/1 to avoid the unfolding rule 
%%% being applied to certain types of implications.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- assert( unfoldable_implication(_ implies _)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
